/*
 * Copyright (c) 2005-2010 Imperas Software Ltd., www.imperas.com
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied.
 *
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */
#include "svdpi.h"
#include "dpiheader.h"

#include "icm/icmCpuManager.h"

static ICM_MEM_WRITE_FN(extMemWriteCB) {

    bfm_write((int)address,*(int *)value);

    icmPrintf(
        "EXTERNAL MEMORY: Writing 0x%08x 0x%08x\n",
         (Int32)address, *(Int32 *)value
    );
}

static ICM_MEM_READ_FN(extMemReadCB) {

    int o;
    bfm_read((int)address,&o);
    *(Int32 *)value = o;

    icmPrintf(
        "EXTERNAL MEMORY: Reading 0x%08x 0x%08x\n",
        (Int32)address, *(Int32 *)value
    );
}


static ICM_MEM_WRITE_FN(watchWriteCB) {

    icmPrintf(
        "WATCHPOINT '%s': Writing to 0x%08x with 0x%08x\n",
	(Uns8 *)userData,
        (Int32)address,
        *(Int32 *)value
    );
}

static ICM_MEM_READ_FN(watchReadCB) {

    icmPrintf(
        "WATCHPOINT '%s': Reading from 0x%08x with 0x%08x\n",
        (Uns8 *)userData,
        (Int32)address,
	*(Int32 *)value
    );
}


icmProcessorP createPlatform(void) {

////////////////////////////////////////////////////////////////////////////////

    icmInit(ICM_VERBOSE|ICM_STOP_ON_CTRLC, 0, 0);

    icmSetPlatformName("BareMetalPowerpc32Single");


////////////////////////////////////////////////////////////////////////////////
//                                 Bus bus                                 //
////////////////////////////////////////////////////////////////////////////////

    icmBusP bus = icmNewBus( "bus", 32);


////////////////////////////////////////////////////////////////////////////////
//                               Processor cpu1                               //
////////////////////////////////////////////////////////////////////////////////
#define SIM_ATTRS (ICM_ATTR_RELAXED_SCHED)

    icmAttrListP cpu1_attr = icmNewAttrList();

    icmAddStringAttr(cpu1_attr, "compatibility", "isa");
    icmAddStringAttr(cpu1_attr, "endian",        "big");
    icmAddDoubleAttr(cpu1_attr, "mips", 100.000000);

    // select library components
    const char *vlnvRoot = NULL; //When NULL use default library
    const char *model       = icmGetVlnvString(vlnvRoot, "ovpworld.org", "processor", "or1k", "1.0", "model");
    const char *semihosting = icmGetVlnvString(vlnvRoot, "ovpworld.org", "modelSupport", "imperasExit", "1.0", "model");

    // create a processor
    icmProcessorP cpu1_c  = icmNewProcessor(
        "cpu1",             // CPU name
        "or1k",             // CPU type
        0,                  // CPU cpuId
        0,                  // CPU model flags
        32,                 // address bits
        model,              // model file
        "modelAttrs",       // morpher attributes
        0,                  // simulation attributes
        0,                  // user-defined attributes
        semihosting,        // semi-hosting file
        "modelAttrs"        // semi-hosting attributes
    );

    icmConnectProcessorBusses( cpu1_c, bus, bus );


////////////////////////////////////////////////////////////////////////////////
//                                Memory memory                               //
////////////////////////////////////////////////////////////////////////////////
    // create two simulated memories for low and high regions
    icmMemoryP memory1 = icmNewMemory("mem1", ICM_PRIV_RWX, 0x003fffff);
    icmMemoryP memory2 = icmNewMemory("mem2", ICM_PRIV_RWX, 0xffffffff-0x00401000);

    // map the address range 0x00400000:0x00400fff externally to the processor,
    // read only
    icmMapExternalMemory(
        bus, "external", 7, 0x00400000, 0x00400fff,
        extMemReadCB, extMemWriteCB, 0
    );

    // connect memories to bus
    icmConnectMemoryToBus(bus, "mp1", memory1, 0);
    icmConnectMemoryToBus(bus, "mp2", memory2, 0x00401000);

    // show the bus connections
    icmPrintBusConnections(bus);
    
    return cpu1_c;
}

////////////////////////////////////////////////////////////////////////////////
//                                   M A I N                                  //
////////////////////////////////////////////////////////////////////////////////

int main() {

 
    const char *appName = "asmtest.OR1K.elf";

    // the constructor
    icmProcessorP processor = createPlatform();

    // load the application executable file into processor memory space
    if(!icmLoadProcessorMemory(processor, appName, False, False, True)) {
        return -1;
    }
/*
    // Create a watchpoint
    // Invoke callback on write accesses to the address 0x00400000-0x00400003
    icmAddWriteCallback(
        processor,          // processor
        0x00400000,         // low address
        0x0040000f,         // high address
        watchWriteCB,       // callback to invoke
        "watch point W"     // user data passed to callback
    );

    // Create a watchpoint
    // Invoke callback on read accesses to the address 0x00400000-0x00400003
    icmAddReadCallback(
        processor,          // processor
        0x00400000,         // low address
        0x0040000f,         // high address
        watchReadCB,        // callback to invoke
        "watch point R"     // user data passed to callback
    );
*/
    icmPrintf("\nStarting Simulation ...\n");

    Bool done = False;

    while(!done) {

        Uns32 currentPC = (Uns32)icmGetPC(processor);

        // disassemble instruction at current PC
        icmPrintf(
            "0x%08x : %s\n", currentPC,
            icmDisassemble(processor, currentPC)
        );

        // execute one instruction
        done = (icmSimulate(processor, 1) != ICM_SR_SCHED);

    }

    icmPrintf("\n");

    // free simulation data structures
    icmTerminate();

    return 0;
}

////////////////////////////////////////////////////////////////////////////////
